unit f_3ds;

interface

const                                       // Konstanty hlaviek jednotlivch blok
  PRIMARY = $4D4D;
  OBJECTINFO = $3D3D;
  VERSION = $0002;
  EDITKEYFRAME = $B000;
  MATERIAL = $AFFF;
  OBJEKT = $4000;
  MATNAME = $A000;
  MATDIFFUSE = $A020;
  MATMAP = $A200;
  MATMAPFILE = $A300;
  OBJECT_MESH = $4100;
  OBJECT_VERTICES = $4110;
  OBJECT_FACES = $4120;
  OBJECT_MATERIAL	=	$4130;
  OBJECT_UV	=	$4140;

type
  CVector3 = record                         // Vektor 3D
    x, y, z: single;
    end;

  CVector2 = record                         // Vektor 2D
    x, y: single;
    end;

  tFace = record                            // Informace o plokch objektu
    vertIndex: array [0..2] of integer;     // Index do pole vertex
    coordIndex: array [0..2] of integer;    // Index do pole texturovch koordint
    end;

  tMaterialInfo = record                    // Informace o materilu
    strName: string;                        // Jmno materilu
    strFile: string;                        // Jmno souboru
    bpc: integer;                           // Poet byt na barvu
    colorub: array [0..2] of byte;          // Barva v bytech
    colorf: array [0..2] of single;         // Barva v singlech
    texureId: integer;                      // ID textury
    uTile: double;                          // Opakovn textury v ose u (nepouito)
    vTile: double;                          // Opakovn textury v ose v (nepouito)
    uOffset: double;                        // Posunut textury v ose u (nepouito)
    vOffset: double;                        // Posunut textury v ose v (nepouito)
    end;

  t3DObject = record                        // Informace o objektu
    numOfVerts: integer;                    // Poet vertex
    numOfFaces: integer;                    // Poet ploek
    numTexVertex: integer;                  // Poet texturovch koordint
    materialID: integer;                    // ID materilu
    bHasTexture: boolean;                   // TRUE, pokud materil obsahuje texturu
    strName: string;                        // Jmno objektu
    pVerts: array of CVector3;              // Vertexy
    pNormals: array of CVector3;            // Normly
    pTexVerts: array of CVector2;           // Texturov koordinty
    pFaces: array of tFace;                 // Ploky
    end;

  Pt3DObject = ^t3DObject;                  // Ukazatel na objekt

  t3DModel = record                         // Informace o modelu
    numOfObjects: integer;                  // Poet objekt
    numOfMaterials: integer;                // Poet materil
    pMaterials: array of tMaterialInfo;     // Pole materil
    pObject: array of t3DObject;            // Pole objekt
    end;

  tChunk = record                           // Informace o bloku
    ID: word;                               // Identifiktor bloku
    length: cardinal;                       // Dlka bloku
    bytesRead: cardinal;                    // Ji peten byty
    end;

  CLoad3DS = class                          // Tda pro nahrn 3DS souboru
    public
      constructor Create;                                                       // Konstruktor
      destructor Destroy; override;                                             // Destruktor
      function Import3DS(var pModel: t3DModel; strFileName: string): boolean;   // Funkce pro nahrn souboru
    private
      m_FilePointer: integer;                                                   // Ukazatel na soubor
      function GetString(var pBuffer: string): integer;                         // Nate etzec
      procedure ReadChunk(var pChunk: tChunk);                                  // Nate dal blok
      procedure ProcessNextChunk(var pModel: t3DModel; var pPreviousChunk: tChunk); // Nate dal soubor blok
      procedure ProcessNextObjectChunk(var pModel: t3DModel; var pObject: t3DObject; var pPreviousChunk: tChunk); // Nate dal objektov blok
      procedure ProcessNextMaterialChunk(var pModel: t3DModel; var pPreviousChunk: tChunk); // Nate dal materilov blok
      procedure ReadColorChunk(var pMaterial: tMaterialInfo; var pChunk: tChunk); // Nate RGB barvu objektu
      procedure ReadVertices(var pObject: t3DObject; var pPreviousChunk: tChunk); // Nate vertexy objektu
      procedure ReadVertexIndices(var pObject: t3DObject; var pPreviousChunk: tChunk);  // Nate ploky objektu
      procedure ReadUVCoordinates(var pObject: t3DObject; var pPreviousChunk: tChunk);  // Nate texturov koordinty objektu
      procedure ReadObjectMaterial(pModel: t3DModel; var pObject: t3DObject; var pPreviousChunk: tChunk); // Nate materil objektu
      procedure ComputeNormals(pModel: t3DModel);                               // Vypot normly
      procedure CleanUp;                                                        // Uvoln prostedky a uzave soubor
    end;

var
  gBuffer: array [0..50000] of integer;                                         // Buffer pro naten nepotebnch dat

implementation

uses SysUtils, Windows;

function Vector(vPoint1, vPoint2: CVector3): CVector3;                          // Vpoet vektoru mezi dvma body
var
  vVector: CVector3;
begin
  vVector.x := vPoint1.x - vPoint2.x;                                           // Bod 1 - Bod 2
	vVector.y := vPoint1.y - vPoint2.y;
	vVector.z := vPoint1.z - vPoint2.z;
  Result := vVector;
end;

function Cross(vVector1, vVector2: CVector3): CVector3;                         // Vektorov souin
var
  vCross: CVector3;
begin
  vCross.x := ((vVector1.y * vVector2.z) - (vVector1.z * vVector2.y));
	vCross.y := ((vVector1.z * vVector2.x) - (vVector1.x * vVector2.z));
	vCross.z := ((vVector1.x * vVector2.y) - (vVector1.y * vVector2.x));
  Result := vCross;
end;

function Normalize(vNormal: CVector3): CVector3;                                // Normalizace vektoru
var
  Magnitude: Double;
begin
  Magnitude := Sqrt(Sqr(vNormal.x) + Sqr(vNormal.y) + Sqr(vNormal.z));          // Velikost vektoru
  vNormal.x := vNormal.x / Magnitude;                                           // Vektor / velikost
	vNormal.y := vNormal.y / Magnitude;
	vNormal.z := vNormal.z / Magnitude;
  Result := vNormal;
end;

function AddVector(vVector1, vVector2: CVector3): CVector3;                     // Souet vektor
var
  vResult: CVector3;
begin
  vResult.x := vVector2.x + vVector1.x;                                         // Vektor 1 + Vektor 2
	vResult.y := vVector2.y + vVector1.y;
	vResult.z := vVector2.z + vVector1.z;
  Result := vResult;
end;

function DivideVectorByScaler(vVector1: CVector3; Scaler: Double): CVector3;    // Dlen vektoru skalrem
var
  vResult: CVector3;
begin
  vResult.x := vVector1.x / Scaler;
	vResult.y := vVector1.y / Scaler;			
	vResult.z := vVector1.z / Scaler;
  Result := vResult;
end;


{ CLoad3DS }

procedure CLoad3DS.CleanUp;                                                     // klid
begin
  if m_FilePointer <> -1 then                                                   // Mme oteven soubor?
    begin
    FileClose(m_FilePointer);                                                   // Zaveme ho
    m_FilePointer := -1;
    end;
end;

procedure CLoad3DS.ComputeNormals(pModel: t3DModel);                            // Vpoet norml
var
  vVector1, vVector2, vNormal: CVector3;
  vPoly: array [0..2] of CVector3;
  index, i, j: integer;
  pObject: Pt3DObject;
  pNormals, pTempNormals: array of CVector3;
  vSum, vZero: CVector3;
  shared: integer;
begin
  if pModel.numOfObjects <= 0 then exit;                                        // Pokud nemme objekt tak konme
  for index := 0 to pModel.numOfObjects - 1 do                                  // Cyklus pes vechny objekty modelu
    begin
    pObject := @pModel.pObject[index];                                          // Zsk aktuln objekt
    SetLength(pNormals,pObject.numOfFaces);                                     // Alokace potebn pamti
    SetLength(pTempNormals,pObject.numOfFaces);
    SetLength(pObject.pNormals,pObject.numOfVerts);
    for i := 0 to pObject.numOfFaces - 1 do                                     // Cyklus pes vechny ploky objektu
      begin
      vPoly[0] := pObject.pVerts[pObject.pFaces[i].vertIndex[0]];               // Pro pehlednost ulo 3 vertexy do pomocn promnn
			vPoly[1] := pObject.pVerts[pObject.pFaces[i].vertIndex[1]];
			vPoly[2] := pObject.pVerts[pObject.pFaces[i].vertIndex[2]];
			// Vlastn vpoet norml
			vVector1 := Vector(vPoly[0], vPoly[2]);		                                // Vpoet 1. vektoru
			vVector2 := Vector(vPoly[2], vPoly[1]);                                   // Vpoet 2. vektoru
			vNormal := Cross(vVector1, vVector2);		                                  // Vpoet normly
			pTempNormals[i] := vNormal;					                                      // Ulo nenormalizovanou normlu pro pozdj vpoty norml vertex
			vNormal := Normalize(vNormal);				                                    // Normalizuje normlu
			pNormals[i] := vNormal;                                                   // Ulo normlu do pole
      end;
    // Vpoet norml vertex
    ZeroMemory(@vSum,sizeof(vSum));
    vZero := vSum;
    shared := 0;
    for i := 0 to pObject.numOfVerts - 1 do                                     // Cyklus pes vechny vertexy
      begin
      for j := 0 to pObject.numOfFaces - 1 do                                   // Cyklus pes vechny ploky (trojhelnky)
        if (pObject.pFaces[j].vertIndex[0] = i) or                              // Je vertex sdlen s jinou plokou?
					  (pObject.pFaces[j].vertIndex[1] = i) or
					  (pObject.pFaces[j].vertIndex[2] = i) then
          begin
          vSum := AddVector(vSum,pTempNormals[j]);                              // Pite nenormalizovanou normlu
          Inc(shared);                                                          // Zv poet sdlench trojhelnk
          end;
      pObject.pNormals[i] := DivideVectorByScaler(vSum,-shared);                // Zsk normlu
      pObject.pNormals[i] := Normalize(pObject.pNormals[i]);                    // Normalizuje normlu
      vSum := vZero;                                                            
      shared := 0;                                                              
      end;
    SetLength(pTempNormals,0);                                                  // Uvoln pam 
    SetLength(pNormals,0);
    end;
end;

constructor CLoad3DS.Create;                                                    // Konstruktor
begin
  m_FilePointer := -1;
end;

destructor CLoad3DS.Destroy;                                                    // Destruktor
begin
  inherited;
end;

function CLoad3DS.GetString(var pBuffer: string): integer;                      // Nate etzec
var
  index: integer;
  tmpChar: Char;
begin
  index := 1;
  FileRead(m_FilePointer,tmpChar,1);                                            // Nate prvn znak
  pBuffer := pBuffer + tmpChar;
  while pBuffer[index] <> #0 do                                                 // Kontroluje, zda jsme na konci etzce (#0)
    begin
    FileRead(m_FilePointer,tmpChar,1);                                          // Nate dal znak
    pBuffer := pBuffer + tmpChar;
    Inc(index);
    end;
 Result := Length(pBuffer);                                                     // Vrac dlku etzce
end;

function CLoad3DS.Import3DS(var pModel: t3DModel;                               // Nahraje model
  strFileName: string): boolean;
var
  strMessage: PAnsiChar;
  currentChunk: tChunk;
begin
  ZeroMemory(@currentChunk,sizeof(currentChunk));
  m_FilePointer := FileOpen(strFileName,fmOpenRead);                            // Oteve 3DS soubor
  if m_FilePointer = -1 then                                                    // Pokud nastala chyba, zobrazme hlku
    begin
    strMessage := PAnsiChar(Format('Unable to find the file: %s!',[strFileName]));
    MessageBox(0,strMessage,'Error',MB_OK);
    Result := false;
    exit;
    end;
  ReadChunk(currentChunk);                                                      // Nate prvn blok
  if currentChunk.ID <> PRIMARY then                                            // Pokud mme jin ne primrn blok, nejedn se o 3DS soubor
    begin
    strMessage := PAnsiChar(Format('Unable to load PRIMARY chuck from file: %s!',[strFileName]));
    MessageBox(0,strMessage,'Error',MB_OK);
    CleanUp;
    Result := false;
    exit;
    end;
  ProcessNextChunk(pModel,currentChunk);                                        // Nate objekty, procedura je volna rekurzivn
  ComputeNormals(pModel);                                                       // Vpoet norml vrchol
  CleanUp;                                                                      // klid
  Result := true;
end;

procedure CLoad3DS.ProcessNextChunk(var pModel: t3DModel;                       // Nate dal bloky
  var pPreviousChunk: tChunk);
var
  currentChunk, tempChunk: tChunk;
begin
  ZeroMemory(@currentChunk,sizeof(currentChunk));
  ZeroMemory(@tempChunk,sizeof(tempChunk));
  while pPreviousChunk.bytesRead < pPreviousChunk.length do                     // Cyklus dokud nenateme cel blok
    begin
    ReadChunk(currentChunk);                                                    // Nate dal podblok
    case currentChunk.ID of                                                     // Vtven podle hlaviky bloku
      VERSION:                                                                  // Informace o verzi souboru
        begin
        currentChunk.bytesRead := currentChunk.bytesRead + FileRead(m_FilePointer,gBuffer,currentChunk.length - currentChunk.bytesRead);
        if (currentChunk.length - currentChunk.bytesRead = 4) and (gBuffer[0] > $03) then   // Pokud je verze vy ne 3, zobrazme varovn
				  MessageBox(0,'This 3DS file is over version 3 so it may load incorrectly','Warning',MB_OK);
        end;
      OBJECTINFO:                                                               // Hlavika objektu
        begin
        ReadChunk(tempChunk);                                                   // Nate dal podblok
        tempChunk.bytesRead := tempChunk.bytesRead + FileRead(m_FilePointer,gBuffer,tempChunk.length - tempChunk.bytesRead);
        currentChunk.bytesRead := currentChunk.bytesRead + tempChunk.bytesRead;
        ProcessNextChunk(pModel,currentChunk);                                  // Nate dal informace o objektu
        end;
      MATERIAL:                                                                 // Informace o materilu
        begin
        Inc(pModel.numOfMaterials);                                             // Zv poet materil
        SetLength(pModel.pMaterials,Length(pModel.pMaterials)+1);
        ProcessNextMaterialChunk(pModel,currentChunk);                          // Nate dal informace o materilu
        end;
      OBJEKT:                                                                   // Informace o objektu
        begin
        Inc(pModel.numOfObjects);                                               // Zv poet objekt
        SetLength(pModel.pObject,Length(pModel.pObject)+1);
        currentChunk.bytesRead := currentChunk.bytesRead + GetString(pModel.pObject[pModel.numOfObjects - 1].strName);  // Nate jmno objektu
        ProcessNextObjectChunk(pModel,pModel.pObject[pModel.numOfObjects-1],currentChunk);  // Nate zbvajc informace o objektu
        end;
      EDITKEYFRAME:                                                             // Klov snmek - nepouito - peskome
        begin
        currentChunk.bytesRead := currentChunk.bytesRead + FileRead(m_FilePointer,gBuffer,currentChunk.length - currentChunk.bytesRead);
        end;
      else                                                                      // Pokud nateme neznm blok, ignorujeme ho
        begin
        currentChunk.bytesRead := currentChunk.bytesRead + FileRead(m_FilePointer,gBuffer,currentChunk.length - currentChunk.bytesRead);
        end;
      end;
    pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + currentChunk.bytesRead;  // Zvtme poet petench byt
    end;
end;

procedure CLoad3DS.ProcessNextMaterialChunk(var pModel: t3DModel;               // Nate informace o materilu
  var pPreviousChunk: tChunk);
var
  currentChunk: tChunk;
begin
  ZeroMemory(@currentChunk,sizeof(currentChunk));
  while pPreviousChunk.bytesRead < pPreviousChunk.length do                     // te dokud nejsme na konci podbloku
    begin
    ReadChunk(currentChunk);                                                    // Nate blok
    case currentChunk.ID of                                                     // Vtven podle hlaviky bloku
      MATNAME:                                                                  // Nate jmno materilu
        currentChunk.bytesRead := currentChunk.bytesRead + GetString(pModel.pMaterials[pModel.numOfMaterials - 1].strName);
      MATDIFFUSE:                                                               // Nate barvu objektu
        ReadColorChunk(pModel.pMaterials[pModel.numOfMaterials - 1],currentChunk);
      MATMAP:                                                                   // Nate informace o materilu
        ProcessNextMaterialChunk(pModel,currentChunk);
      MATMAPFILE:                                                               // Nate jmno souboru s texturou
        currentChunk.bytesRead := currentChunk.bytesRead + GetString(pModel.pMaterials[pModel.numOfMaterials - 1].strFile);
      else                                                                      // Ignorujeme neznm bloky
      currentChunk.bytesRead := currentChunk.bytesRead + FileRead(m_FilePointer,gBuffer,currentChunk.length - currentChunk.bytesRead);
      end;
    pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + currentChunk.bytesRead;  // Zvtme poet petench byt
    end;
end;

procedure CLoad3DS.ProcessNextObjectChunk(var pModel: t3DModel;                 // Nate informace o objektu
  var pObject: t3DObject; var pPreviousChunk: tChunk);
var
  currentChunk: tChunk;
begin
  while pPreviousChunk.bytesRead < pPreviousChunk.length do                     // te do konce podbloku
    begin
    ZeroMemory(@currentChunk,sizeof(currentChunk));
    ReadChunk(currentChunk);                                                    // Nate dal blok
    case currentChunk.ID of                                                     // Vtven podle hlaviky
      OBJECT_MESH:                                                              // Nov objekt
        ProcessNextObjectChunk(pModel,pObject,currentChunk);                    // Nate jeho informace
      OBJECT_VERTICES:                                                          // Nate vertexy objektu
        ReadVertices(pObject,currentChunk);
      OBJECT_FACES:                                                             // Nate ploky objektu
        ReadVertexIndices(pObject,currentChunk);
      OBJECT_MATERIAL:                                                          // Nate jmno pouitho materilu
        ReadObjectMaterial(pModel,pObject,currentChunk);
      OBJECT_UV:                                                                // Nate texturov koordinty objektu
        ReadUVCoordinates(pObject,currentChunk);
      else                                                                      // Ignorujeme neznm bloky
      currentChunk.bytesRead := currentChunk.bytesRead + FileRead(m_FilePointer,gBuffer,currentChunk.length - currentChunk.bytesRead);
      end;
    pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + currentChunk.bytesRead;  // Zvtme poet petench byt
    end;
end;

procedure CLoad3DS.ReadColorChunk(var pMaterial: tMaterialInfo;                 // Nate barvu objektu
  var pChunk: tChunk);
var
  tempChunk: tChunk;
begin
  ZeroMemory(@tempChunk,sizeof(tempChunk));
  ReadChunk(tempChunk);                                                         // Informace o bloku
  pMaterial.bpc := tempChunk.length - tempChunk.bytesRead;
  if pMaterial.bpc = 3 then                                                     // Podle dlky bloku nateme barvu do pslunho pole (3-byte, 12-single)
    tempChunk.bytesRead := tempChunk.bytesRead + FileRead(m_FilePointer,pMaterial.colorub,pMaterial.bpc)
    else
    tempChunk.bytesRead := tempChunk.bytesRead + FileRead(m_FilePointer,pMaterial.colorf,pMaterial.bpc);
  pChunk.bytesRead := pChunk.bytesRead + tempChunk.bytesRead;                   // Zvtme poet petench byt
end;

procedure CLoad3DS.ReadChunk(var pChunk: tChunk);                               // Nate hlaviku bloku
begin
  pChunk.bytesRead := FileRead(m_FilePointer,pChunk.ID,2);                      // ID bloku
  pChunk.bytesRead := pChunk.bytesRead + FileRead(m_FilePointer,pChunk.length,4); // Dlka bloku
end;

procedure CLoad3DS.ReadObjectMaterial(pModel: t3DModel;                         // Nate materily
  var pObject: t3DObject; var pPreviousChunk: tChunk);
var
  strMaterial: string;
  i: integer;
begin
  strMaterial := '';
  pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + GetString(strMaterial);// Jmno materilu
  for i := 0 to pModel.numOfMaterials - 1 do                                    // Projde vechny materily
    if strMaterial = pModel.pMaterials[i].strName then                          // Nali jsme n materil?
      begin
      pObject.materialID := i;                                                  // Nastavme jeho index
      if pModel.pMaterials[i].strFile <> '' then                                // Pokud existuje soubor s texturou
        pObject.bHasTexture := true;                                            // Nastavme pznak
      break;
      end
      else
      pObject.materialID := -1;                                                 // Objekt nem materil
  pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + FileRead(m_FilePointer,gBuffer,pPreviousChunk.length - pPreviousChunk.bytesRead);  // Zvtme poet petench byt
end;

procedure CLoad3DS.ReadUVCoordinates(var pObject: t3DObject;                    // Nate texturov koordinty
  var pPreviousChunk: tChunk);
var
  i: integer;
begin
  pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + FileRead(m_FilePointer,pObject.numTexVertex,2);  // Nate poet koordint
  SetLength(pObject.pTexVerts,pObject.numTexVertex);                            // Nastav dlku pole
  for i := 0 to pObject.numTexVertex - 1 do                                     // Nate vechny koordinty
    pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + FileRead(m_FilePointer,pObject.pTexVerts[i],8);
end;

procedure CLoad3DS.ReadVertexIndices(var pObject: t3DObject;                    // Nate indexy do pole vertex
  var pPreviousChunk: tChunk);
var
  index: word;
  i, j: integer;
begin
  pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + FileRead(m_FilePointer,pObject.numOfFaces,2);  // Nate poet ploek
  SetLength(pObject.pFaces,pObject.numOfFaces);                                 // Nastav velikost pole
  ZeroMemory(pObject.pFaces,pObject.numOfFaces);
  for i := 0 to pObject.numOfFaces - 1 do                                       // Cyklus pes vechny ploky
    for j := 0 to 3 do                                                          // Cyklus pes vrcholy ploek
      begin
      pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + FileRead(m_FilePointer,index,sizeof(index)); // Nate index vrcholu
      if j < 3 then
        pObject.pFaces[i].vertIndex[j] := index;                                // Ulo index do pole
      end;
end;

procedure CLoad3DS.ReadVertices(var pObject: t3DObject;                         // Nate vertexy
  var pPreviousChunk: tChunk);
var
  i: integer;
  fTempY: Double;
begin
  pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + FileRead(m_FilePointer,pObject.numOfVerts,2);  // Nate poet vertex
  SetLength(pObject.pVerts,pObject.numOfVerts);                                 // Nastav velikost pole
  ZeroMemory(pObject.pVerts,pObject.numOfVerts);
  for i := 0 to pObject.numOfVerts - 1 do                                       // Cyklus pes vechny vertexy
    begin
    pPreviousChunk.bytesRead := pPreviousChunk.bytesRead + FileRead(m_FilePointer,pObject.pVerts[i],12);
    fTempY := pObject.pVerts[i].y;                                              // Prohod Y a Z
    pObject.pVerts[i].y := pObject.pVerts[i].z;
    pObject.pVerts[i].z := -fTempY;                                             // Jet zmna orientace osy
    end;
end;

end.
