program main;

{   kd pro Delphi 7}

uses
  Windows,
  SysUtils,
  Messages,
  OpenGL,
  glaux,
  init in 'init.pas',
  f_3ds in 'f_3ds.pas';

procedure glGenTextures(n: GLsizei; var textures: GLuint); stdcall; external opengl32;
procedure glBindTexture(target: GLenum; texture: GLuint); stdcall; external opengl32;
procedure glDeleteTextures(n: GLsizei; textures: PGLuint); stdcall; external opengl32;

var
  Title: string = '3ds loader';                        // Titulek okna
  ResWidth: integer = 800;                             // ka okna
  ResHeight: integer = 600;                            // Vka okna
  ResBPP: integer = 16;                                // Barevn hloubka
                                                       // Vdy nechte odkomentovan pouze jeden dek, ostatn zakomentujte!!!
  FILE_NAME: string = 'face.3ds';                      // Soubor, kter budeme nahrvat - originl
  //FILE_NAME: string = 'muz.3ds';                     // Pedpipraven objekt v Cinema 4D
  //FILE_NAME: string = 'snehulak.3ds';                // Mj vtvor

  g_window: PGL_Window;                                 // Okno
  g_keys: PKeys;                                        // Klvesy
  g_Texture: array of UINT;                             // Pole textur
  g_Load3ds: CLoad3DS;                                  // Tda 3DS
  g_3DModel: t3DModel;                                  // 3DS model
  g_ViewMode: integer = GL_TRIANGLES;                   // Zpsob vykreslovn
  g_bLighting: boolean = true;                          // Osvtlen
  g_RotateX: GLfloat = 0.0;                             // Rotace
  g_RotationSpeed: GLfloat = 0.8;                       // Rychlost rotace



procedure CreateTexture(var textureArray: array of UINT; strFileName: LPSTR; textureID: integer);       // Vytvoen textury
var
  pBitmap: PTAUX_RGBImageRec;                                                   // Pomocn promnn pro nahrn bitmapy
begin
  if strFileName = '' then exit;                                                // Bylo pedno jmno soubor?
  pBitmap := auxDIBImageLoadA(strFileName);                                     // Naten bitmapy
  if pBitmap = nil then exit;                                                   // Podailo se nast bitmapu?
  glGenTextures(1,textureArray[textureID]);                                     // Generujeme 1 texturu
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);                                         // Zpsob interpretace dat v pamti (1 - hodnoty typu byte)
  glBindTexture(GL_TEXTURE_2D,textureArray[textureID]);                         // Zvol texturu
  gluBuild2DMipmaps(GL_TEXTURE_2D,3,pBitmap.sizeX,pBitmap.sizeY,GL_RGB,GL_UNSIGNED_BYTE,pBitmap.data);  // Vytvo mipmapovou texturu
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);// Zpsob filtrovn
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR_MIPMAP_LINEAR);
end;

function Initialize(window: PGL_Window; key: PKeys): boolean;	                  // Inicializace OpenGL
var
  i: integer;                                                                   // Cyklus
begin
  g_window := window;
  g_keys := key;
  if ParamCount <> 0 then FILE_NAME := ParamStr(1);                             // Pokud byl programu pedn soubor jako parametr, pak jej nateme msto souboru, kter je definovm pmo v programu
  ZeroMemory(@g_3DModel,sizeof(g_3DModel));                                     // Nulovn pamti
  g_Load3ds := CLoad3DS.Create;                                                 // Vytvo instanci tdy 3DS
  g_Load3ds.Import3DS(g_3DModel,FILE_NAME);                                     // Nahraje model
  SetLength(g_Texture,g_3DModel.numOfMaterials);                                // Nastav velikost pole pro textury
  for i := 0 to g_3DModel.numOfMaterials -1 do                                  // Cyklus pes jednotliv materily
    begin
    if g_3DModel.pMaterials[i].strFile <> '' then                               // Obsahuje materil texturu?
      CreateTexture(g_Texture,PChar(g_3DModel.pMaterials[i].strFile),i);        // Pokud ano, pak ji vytvome
    g_3DModel.pMaterials[i].texureId := i;                                      // Ulo ID textury
    end;
  glEnable(GL_LIGHT0);								                                          // Zapne svtlo
	glEnable(GL_LIGHTING);								                                        // Zapne osvtlen
	glEnable(GL_COLOR_MATERIAL);                                                  // Povol barvu materilu
  glEnable(GL_TEXTURE_2D);							                                        // Povol mapovn textur
	glEnable(GL_DEPTH_TEST);                                                      // Zapne testovn hloubky
  Result:=true;                                                                 // Inicializace probhla v podku
end;

procedure Deinitialize;                                                         // Uvolnn prostedk
var
  i: integer;                                                                   // Cyklus
begin
  for i := 0 to g_3DModel.numOfObjects - 1 do                                   // Projde vechny objekty v modelu
    begin
    SetLength(g_3DModel.pObject[i].pVerts,0);                                   // Uvoln pole vertex
    SetLength(g_3DModel.pObject[i].pNormals,0);                                 // Uvoln pole norml
    SetLength(g_3DModel.pObject[i].pTexVerts,0);                                // Uvoln pole texturovanch vertex
    SetLength(g_3DModel.pObject[i].pFaces,0);                                   // Uvoln pole fac
    end;
  glDeleteTextures(g_3DModel.numOfMaterials,@g_Texture);                        // Uvoln pam pro textury
  SetLength(g_Texture,0);                                                       // Uvoln pole textur
  g_Load3ds.Free;                                                               // Zru instanci tdy 3DS
end;

procedure Update(milliseconds: DWORD);                                // Aktualizace pohyb ve scn a stisk klves
begin
  if g_keys.keyDown[VK_ESCAPE] then                                   // Klvesa ESC?
    TerminateApplication(g_window^);                                  // Ukonen programu
  if g_keys.keyDown[VK_F1] then                                       // Klvesa F1?
    ToggleFullscreen(g_window^);                                      // Pepnut fullscreen/okno
  if g_keys.keyDown[VK_LEFT] then                                     // ipka vlevo
    g_RotationSpeed := g_RotationSpeed - 0.05;                        // prava rotace
  if g_keys.keyDown[VK_RIGHT] then                                    // ipka vpravo
    g_RotationSpeed := g_RotationSpeed + 0.05;                        // prava rotace
end;

procedure Draw;                                                                 // Vykreslen scny
var
  i, j, whichVertex: integer;                                                   // Cykly
  pObject: t3DObject;                                                           // Objekt
  index: integer;                                                               // Index do pole vertex
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);                          // Smae obrazovku a hloubkov buffer
  glLoadIdentity;                                                               // Reset matice
  gluLookAt(0,1.5,8,0,0.5,0,0,1,0);                                             // Nastaven pohledu kamery
  glRotatef(g_RotateX,0,1.0,0);                                                 // Rotace modelu
  g_RotateX := g_RotateX + g_RotationSpeed;                                     // Aktualizace hlu natoen
  for i := 0 to g_3DModel.numOfObjects - 1 do                                   // Projde vechny objekty v modelu
    begin
   if Length(g_3DModel.pObject) = 0 then break;                                 // Pokud model neobsahuje dn objekt, ukonme cyklus
    pObject := g_3DModel.pObject[i];                                            // Ulome objekt do pomocn promnn
    if pObject.bHasTexture then                                                 // M objekt texturu?
      begin                                                                     // Pokud ano
      glEnable(GL_TEXTURE_2D);                                                  // Zapne mapovn textur
      glColor3ub(255,255,255);
      glBindTexture(GL_TEXTURE_2D,g_Texture[pObject.materialID]);               // Zvol pslunou texturu
      end
      else                                                                      // Pokud ne
      begin
      glDisable(GL_TEXTURE_2D);                                                 // Vypne mapovn textur
      glColor3ub(255,255,255);
      end;
    glBegin(g_ViewMode);                                                        // Zatek vykreslovn vertex
      for j := 0 to pObject.numOfFaces - 1 do                                   // Projde vechny trojhelnky
        for whichVertex := 0 to 2 do                                            // Projde vechny vrcholy trojhelnk
          begin
          index := pObject.pFaces[j].vertIndex[whichVertex];                    // Index do pole vertex
          glNormal3f(pObject.pNormals[index].x,pObject.pNormals[index].y,pObject.pNormals[index].z);  // Nastav normlu
          if pObject.bHasTexture then                                           // M objekt texturu?
            begin
            if Assigned(pObject.pTexVerts) then                                 // M objekt texturov koordinty?
              glTexCoord2f(pObject.pTexVerts[index].x,pObject.pTexVerts[index].y);  // Nastav texturov koordinty
            end
            else
            begin
            if (Length(g_3DModel.pMaterials) <> 0) and (pObject.materialID >= 0) then // Kdy objekt nem texturu m alespo materil?
              if g_3DModel.pMaterials[pObject.materialID].bpc = 3 then          // Podle potu byt na barevn kanl, pouijeme vhodnou funkci pro nastaven barvy
                glColor3ubv(@g_3DModel.pMaterials[pObject.materialID].colorub)
                else
                glColor3fv(@g_3DModel.pMaterials[pObject.materialID].colorf);
            end;
          glVertex3f(pObject.pVerts[index].x,pObject.pVerts[index].y,pObject.pVerts[index].z);  // Nakonec vykreslme vertex
          end;
    glEnd;                                                                      // Konec vykreslovn vertex
    end;
	glFlush;                                                                      // Vyprzdn OpenGL renderovac pipeline
end;

function WindowProc(hWnd: HWND;                                       // Handle okna
                 uMsg: UINT;                                          // Zprva pro okno
                 wParam: WPARAM;                                      // Doplkov informace
                 lParam: LPARAM):                                     // Doplkov informace
                                  LRESULT; stdcall;
var
  window: ^GL_Window;
  creation: ^CREATESTRUCT;
begin
  if uMsg = WM_SYSCOMMAND then                                        // Systmov pkaz
      case wParam of                                                  // Typ systmovho pkazu
        SC_SCREENSAVE,SC_MONITORPOWER:                                // Pokus o zapnut etie obrazovky, Pokus o pechod do spornho reimu?
          begin
            Result := 0;                                              // Zabrn obojmu
            exit;
          end;
      end;
  window := Pointer(GetWindowLong(hWnd,GWL_USERDATA));
  case uMsg of                                                        // Vtven podle pchoz zprvy
    WM_CREATE:                                                        // Vytvoen okna
      begin
      creation := Pointer(lParam);
      window := Pointer(creation.lpCreateParams);
      SetWindowLong(hWnd,GWL_USERDATA,Integer(window));
      Result := 0;
      end;
    WM_CLOSE:                                                         // Povel k ukonen programu
      begin
      TerminateApplication(window^);                                  // Pole zprvu o ukonen
      Result := 0                                                     // Nvrat do hlavnho cyklu programu
      end;
    WM_SIZE:                                                          // Zmna velikosti okna
      begin
      case wParam of
        SIZE_MINIMIZED:
          begin
          window.isVisible := false;
          Result := 0;
          end;
        SIZE_MAXIMIZED,
        SIZE_RESTORED:
          begin
          window.isVisible := true;
          ReshapeGL(LOWORD(lParam),HIWORD(lParam));                 // LoWord=ka, HiWord=Vka
          Result := 0;                                              // Nvrat do hlavnho cyklu programu
          end;
      end;
      end;
    WM_KEYDOWN:                                                     // Stisk klvesy
      begin
      if (wParam >= 0) and (wParam <= 255) then
        begin
        window^.keys^.keyDown[wParam] := true;                      // Oznm to programu
        Result := 0;                                                // Nvrat do hlavnho cyklu programu
        end;
      end;
    WM_KEYUP:                                                       // Uvolnn klvesy
      begin
      if (wParam >= 0) and (wParam <= 255) then
        begin
    	  window^.keys^.keyDown[wParam] := false;                     // Oznm to programu
        Result := 0;                                                // Nvrat do hlavnho cyklu programu
        end;
      end;
    WM_TOGGLEFULLSCREEN:                                            // Pepnut fullscreen/okno
      begin
      g_createFullScreen := not g_createFullScreen;
      PostMessage(hWnd,WM_QUIT,0,0);
      Result := 0;
      end;
    WM_LBUTTONDOWN:                                                 // Obsluha levho tlatka myi
      begin
      if g_ViewMode = GL_TRIANGLES then
        g_ViewMode := GL_LINE_STRIP                                 // Zmna mdu vykreslovn
        else
        g_ViewMode := GL_TRIANGLES;
      Result := 0;
      end;
    WM_RBUTTONDOWN:                                                 // Obsluha pravho tlatka myi
      begin
      g_bLighting := not g_bLighting;                               // Zapne/vypne osvtlen
      if g_bLighting then
        glEnable(GL_LIGHTING)
        else
        glDisable(GL_LIGHTING);
      Result := 0;
      end;
    else
      // Pedn ostatnch zprv systmu
      begin
      	Result := DefWindowProc(hWnd,uMsg,wParam,lParam);
      end;
    end;
end;

function RegisterWindowClass(application: Application): boolean;
var
  windowClass: WNDCLASSEX;
begin
  ZeroMemory(@windowClass,Sizeof(windowClass));
  with windowClass do
    begin
    cbSize := Sizeof(windowClass);
    style := CS_HREDRAW or CS_VREDRAW or CS_OWNDC;                  // Pekreslen pi zmn velikosti a vlastn DC
    lpfnWndProc := @WindowProc;                                     // Definuje proceduru okna
    hInstance := application.hInstance;                             // Instance
    hbrBackground := COLOR_APPWORKSPACE;                            // Pozad nen nutn
    hCursor := LoadCursor(0,IDC_ARROW);                             // Standardn kurzor myi
    lpszClassName := PChar(application.className);                  // Jmno tdy okna
    end;
  if RegisterClassEx(windowClass) = 0 then                          // Registruje tdu okna
    begin
    MessageBox(HWND_DESKTOP,'RegisterClassEx Failed!','Error',MB_OK or MB_ICONEXCLAMATION);
    Result := false;                                                // Pi chyb vrt false
    exit;
    end;
  Result := true;
end;

function WinMain(hInstance: HINST;                                  // Instance
		 hPrevInstance: HINST;                                          // Pedchoz instance
		 lpCmdLine: PChar;                                              // Parametry pkazov dky
		 nCmdShow: integer):                                            // Stav zobrazen okna
                        integer; stdcall;
var
  app: Application;
  window: GL_Window;
  key: Keys;
  isMessagePumpActive: boolean;
  msg: TMsg;                                                        // Struktura zprv systmu
  tickCount: DWORD;
begin
  app.className := 'OpenGL';
  app.hInstance := hInstance;
  ZeroMemory(@window,Sizeof(window));
  with window do
    begin
    keys := @key;
    init.application := @app;
    init.title := Title;
    init.width := ResWidth;
    init.height := ResHeight;
    init.bitsPerPixel := ResBPP;
    init.isFullScreen := true;
    end;
  ZeroMemory(@key,Sizeof(key));
  // Dotaz na uivatele pro fullscreen/okno
  if MessageBox(HWND_DESKTOP,'Would You Like To Run In FullScreen Mode?','Start FullScreen',
                MB_YESNO or MB_ICONQUESTION) = IDNO then
    window.init.isFullScreen := false;                                 // Bh v okn
  if not RegisterWindowClass(app) then
    begin
    MessageBox(HWND_DESKTOP,'Error Registering Window Class!','Error',MB_OK or MB_ICONEXCLAMATION);
    Result := -1;
    exit;
    end;
  g_isProgramLooping := true;
  g_createFullScreen := window.init.isFullScreen;
  while g_isProgramLooping do
    begin
    window.init.isFullScreen := g_createFullScreen;
    if CreateWindowGL(window) then
      begin
      if not Initialize(@window,@key) then
        TerminateApplication(window)
        else
        begin
        isMessagePumpActive := true;
        while isMessagePumpActive do
          if PeekMessage(msg,0,0,0,PM_REMOVE) then                  // Pila zprva?
            if msg.message <> WM_QUIT then                          // Obdreli jsme zprvu pro ukonen?
              DispatchMessage(msg)                                
              else
              isMessagePumpActive := false                          // Konec programu
            else
            if not window.isVisible then
              WaitMessage
              else
              begin
              tickCount := GetTickCount;
              Update(tickCount - window.lastTickCount);
              window.lastTickCount := tickCount;
              Draw;
              SwapBuffers(window.hDc);
              end;
        end;
      Deinitialize;
      DestroyWindowGL(window);
      end
      else
      begin
      MessageBox(HWND_DESKTOP,'Error Creating OpenGL Window','Error',MB_OK or MB_ICONEXCLAMATION);
      g_isProgramLooping := false;
      end;
    end;
  UnregisterClass(PChar(app.className),app.hInstance);
  Result := 0;
end;

begin
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow );                  // Start programu
end.


